import math
import operator

from django import forms
from django.db import models
from django.db.models.fields import files

from .storage import VimeoFileStorage


class VimeoFieldFile(files.FieldFile):
    def get_meta(self):
        return self.storage.get_meta(self.name)
    meta = property(get_meta)

    def get_oembed(self, **options):
        return self.storage.get_oembed(self.name, **options)
    oembed = property(get_oembed)

    def get_embed_code(self, **options):
        return self.storage.get_embed_code(self.name, **options)

    def _get_optimal_index(self, iterable, width=None, height=None):
        sizes = [(i, r.get('width', 0), r.get('height', 0)) for i, r in enumerate(iterable)]
        if width and height:
            distances = (
                (r[0], math.sqrt(math.pow(r[1] - width, 2) + math.pow(r[2] - height, 2))) for r in sizes
            )
            return sorted(distances, key=operator.itemgetter(1))[0][0]
        elif width:
            key = 1
            val = width
        elif height:
            key = 2
            val = height
        distances = (
            (r[0], math.pow(val - r[key], 2)) for r in sizes
        )
        return sorted(distances, key=operator.itemgetter(1))[0][0]

    def get_optimal_file(self, width=None, height=None):
        iterable = self.meta.get('files', [])
        return iterable[self._get_optimal_index(iterable, width, height)] if iterable else None

    def get_optimal_picture(self, width=None, height=None):
        pictures = self.meta.get('pictures', {})
        if not pictures:
            return None
        iterable = pictures.get('sizes', [])
        return iterable[self._get_optimal_index(iterable, width, height)] if iterable else None

    def get_optimal_download(self, width=None, height=None):
        iterable = self.meta.get('download', [])
        return iterable[self._get_optimal_index(iterable, width, height)] if iterable else None


class VimeoField(models.FileField):
    """
    Model field for vimeo video. Descendant of
    :py:class:`django.db.models.FileField`.
    """
    attr_class = VimeoFieldFile

    def __init__(self, *args, **kwargs):
        defaults = {'storage': VimeoFileStorage()}
        defaults.update(kwargs)
        super(VimeoField, self).__init__(*args, **defaults)

    def formfield(self, **kwargs):
        defaults = {'form_class': VimeoFormField}
        defaults.update(kwargs)
        return super(VimeoField, self).formfield(**defaults)

    def deconstruct(self):
        name, path, args, kwargs = super(VimeoField, self).deconstruct()
        return name, path, args, kwargs

    def south_field_triple(self):
        name, path, args, kwargs = self.deconstruct()
        return '{}.{}'.format(path, name), args, kwargs


class VimeoFormField(forms.FileField):
    """
    Form field for vimeo video. Descendant of
    :py:class:`django.forms.FileField`
    """
    pass



import logging
import re

import requests
from django.template import Library, Node
from django.utils.encoding import smart_str
from django.utils.safestring import mark_safe

from .. import exceptions

register = Library()

logger = logging.getLogger(__name__)


@register.tag('vimeo')
class VimeoNode(Node):
    """
    Template tag ``vimeo``. It gives access to all
    Keys: https://developer.vimeo.com/apis/oembed#arguments

    Usage (shortcut):
    .. code-block:: html+django
        {% vimeo instance.vimeo_uri [key1=value1, key2=value2...] %}
    Or as a block:
    .. code-block:: html+django
        {% vimeo instance.vimeo_uri [key1=value1, key2=value2...] as VAR %}
            ...
        {% endvideo %}
    Examples:
    .. code-block:: html+django
        {% vimeo instance.vimeo_uri %}
        {% vimeo instance.vimeo_uri width=600 %}
        {% vimeo instance.vimeo_uri autoplay=True loop=True as my_video %}
            HTML: {{ my_video.html }}
            Thumbnail: {{ my_video.thumbnail_url }}
        {% endvideo %}
    """
    error_msg = 'Syntax error. Expected: ``{% video instance.vimeo_uri ' \
                '[key1=val1 key2=val2 ...] [as var] %}``'

    re_option = re.compile(r'^(?P<key>[\w]+)=(?P<value>.+)$')

    def __init__(self, parser, token):
        """
        :param parser: Django template parser
        :type parser: django.template.base.Parser
        :param token: Django template token
        :type token: django.template.base.Token
        """
        self.parser = parser
        self.bits = list(token.split_contents())
        self.tag_name = str(self.pop_bit())
        self.video = self.pop_bit()

        if len(self.bits) > 1 and self.bits[-2] == 'as':
            del self.bits[-2]
            self.variable_name = str(self.pop_bit(-1))
            self.nodelist_file = parser.parse(('end' + self.tag_name, ))
            parser.delete_first_token()
        else:
            self.variable_name = None

        self.options = self.parse_options(self.bits)

    def pop_bit(self, index=0):
        return self.parser.compile_filter(self.bits.pop(index))

    def parse_options(self, bits):
        options = {}
        for bit in bits:
            parsed_bit = self.re_option.match(bit)
            key = smart_str(parsed_bit.group('key'))
            value = self.parser.compile_filter(parsed_bit.group('value'))
            options[key] = value
        return options

    def render(self, context):
        """
        Returns generated HTML.
        :param context: Django template RequestContext
        :type context: django.template.RequestContext
        :return: Rendered HTML with embed video.
        :rtype: django.utils.safestring.SafeText | str
        """
        video = self.video.resolve(context)
        options = self.resolve_options(context)
        try:
            if not self.variable_name:
                return self.embed(video, context=context, **options)
            video_meta = video.meta
            width, height = options.get('width'), options.get('height')
            video_meta.update(
                    {'optimal_file': video.get_optimal_file(width, height),
                     'optimal_picture': video.get_optimal_picture(width, height),
                     'optimal_download': video.get_optimal_download(width, height),
                     'oembed': video.get_oembed(**options)})
            return self.render_block(context, video_meta)
        except requests.Timeout:
            logger.exception('Timeout reached during rendering embed video (`{0}`)'.format(video))
        except exceptions.VideoDoesntExistException:
            logger.exception('Attempt to render not existing video (`{0}`)'.format(video))

        return ''

    def resolve_options(self, context):
        """
        :param context: Django template RequestContext
        :type context: django.template.RequestContext
        """
        options = {}
        for key in self.options:
            value = self.options[key]
            options[key] = value.resolve(context)
        return options

    def render_block(self, context, data):
        """
        :param context: Django template RequestContext
        :type context: django.template.RequestContext
        :param backend: Given instance inherited from VideoBackend
        :type backend: VideoBackend
        :rtype: django.utils.safestring.SafeText
        """
        context.push()
        context[self.variable_name] = data
        output = self.nodelist_file.render(context)
        context.pop()
        return output

    @classmethod
    def embed(cls, video, context=None, **options):
        """
        Direct render of embed video.
        :param video: video
        :type video: VimeoFieldFile
        :param context: Django template RequestContext
        :type context: django.template.RequestContext | None
        """
        return mark_safe(video.get_embed_code(**options))

    def __iter__(self):
        for node in self.nodelist_file:
            yield node

    def __repr__(self):
        return '<VimeoNode "%s">' % self.url


class VimeoException(Exception):
    """ Parental class for all django_vimeo exceptions """
    pass


class VideoDoesntExistException(VimeoException):
    """ Exception thrown if video doesn't exist """
    pass


class UnknownIdException(VideoDoesntExistException):
    """
    Exception thrown if backend is detected, but video ID cannot be parsed.
    """
    pass


class SpaceNotEnoughtException(VimeoException):
    """
    Exception thrown if backend is detected, but video ID cannot be parsed.
    """
    pass


import datetime
import operator
import os
from tempfile import NamedTemporaryFile

import requests
from django.conf import settings
from django.core.files.storage import Storage
from django.utils.deconstruct import deconstructible
from django.utils.functional import cached_property
from vimeo import VimeoClient

from .cache import cache_it
from .exceptions import SpaceNotEnoughtException
from .exceptions import UnknownIdException


@deconstructible
class VimeoFileStorage(Storage):
    def __init__(self, client_id=None, client_secret=None, access_token=None):
        if not client_id:
            client_id = getattr(settings, 'VIMEO_CLIENT_ID', None)
        if not client_secret:
            client_secret = getattr(settings, 'VIMEO_CLIENT_SECRET', None)
        if not access_token:
            access_token = getattr(settings, 'VIMEO_ACCESS_TOKEN', None)
        self.client_id = client_id
        self.client_secret = client_secret
        self.access_token = access_token
        self.oembed_url = getattr(settings, 'VIMEO_OEMBED_URL',
                                  'https://vimeo.com/api/oembed.json')
        self.video_url_pattern = getattr(settings, 'VIMEO_VIDEO_URL_PATTERN',
                                         'https://vimeo.com/{}')

    @cached_property
    def client(self):
        return VimeoClient(token=self.access_token, key=self.client_id, secret=self.client_secret)

    def _upload_quota_check(self, size):
        res = self.client.get('/me')
        self._raise_for_status(res)
        res = res.json()
        if 'upload_quota' in res and \
                        res.get('upload_quota', {}).get('space', {}).get('free', 0) < size:
            freeMb = res.get('upload_quota', {}).get('space', {}).get('free', 0) / 1048576
            usedMb = res.get('upload_quota', {}).get('space', {}).get('used', 0) / 1048576
            maxMb = res.get('upload_quota', {}).get('space', {})('max', 0) / 1048576
            sizeMb = size / 1048576
            raise SpaceNotEnoughtException(
                'Space on vimeo not enought for {} MB (free: {} MB, used: {} MB, max: {} MB)'
                    .format(sizeMb, freeMb, usedMb, maxMb)
            )

    def _raise_for_status(self, res):
        if res.status_code == 404:
            raise UnknownIdException(res.reason)
        res.raise_for_status()

    @cache_it()
    def get_meta(self, path):
        res = self.client.get(path)
        self._raise_for_status(res)
        return res.json()

    @cache_it()
    def get_oembed(self, path, **options):
        defaults = {'url': self.video_url_pattern.format(path.split('/')[-1])}
        defaults.update(options)
        res = requests.get(self.oembed_url, params=defaults)
        self._raise_for_status(res)
        return res.json()

    def get_embed_code(self, path, **options):
        oembed = self.get_oembed(path, **options)
        return oembed.get('html')

    def get_valid_name(self, name):
        valid_name = name
        if valid_name and valid_name[0] == '.':
            valid_name = valid_name[1:]
        return valid_name

    def get_available_name(self, name, max_length=None):
        return self.get_valid_name(name)

    def path(self, name):
        return name

    def delete(self, name):
        self.client.delete(name)

    def exists(self, name):
        return self.client.get(name).ok

    def size(self, name):
        res = self.client.get(name)
        self._raise_for_status(res)
        res = res.json()
        large_files = sorted(res.get('files'), key=operator.itemgetter('size'), reverse=True)
        return large_files[0].get('size', 0) if large_files else 0

    def url(self, name):
        res = self.client.get(name)
        res.raise_for_status()
        res = res.json()
        large_files = sorted(res.get('files'), key=operator.itemgetter('size'), reverse=True)
        return large_files[0].get('link_secure') if large_files else None

    def accessed_time(self, name):
        res = self.client.get(name)
        res.raise_for_status()
        res = res.json()
        utc_datetime = res.get('modified_time')
        return datetime.strptime(utc_datetime, '%Y-%m-%dT%H:%M:%SZ')

    def created_time(self, name):
        res = self.client.get(name)
        res.raise_for_status()
        res = res.json()
        utc_datetime = res.get('created_time')
        return datetime.strptime(utc_datetime, '%Y-%m-%dT%H:%M:%SZ')

    def modified_time(self, name):
        return self.accessed_time(name)

    def _save(self, name, content):
        size = content.size
        self._upload_quota_check(size)
        uploaded_uri = None
        if hasattr(content, 'temporary_file_path'):
            tmp_file_path = content.temporary_file_path()
            uploaded_uri = self.client.upload(tmp_file_path)
        else:
            try:
                tmp_file = NamedTemporaryFile(delete=False)
                for chunk in content.chunks():
                    if not chunk:
                        continue
                    tmp_file.write(chunk)
                tmp_file.close()
                uploaded_uri = self.client.upload(tmp_file.name)
            finally:
                os.unlink(tmp_file.name)
        return uploaded_uri



from setuptools import setup, find_packages

import os

django_vimeo = __import__('django_vimeo')


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

README = read('README.rst')
CHANGES = read('CHANGES.rst')

setup(
    name='django-vimeo',
    packages=find_packages(),
    version=django_vimeo.get_version(),
    author='Georgy Kutsurua',
    author_email='g.kutsurua@gmail.com',
    url='https://github.com/suquant/django-vimeo',
    description=django_vimeo.__doc__.strip(),
    long_description='\n\n'.join([README, CHANGES]),
    classifiers=[
        'Framework :: Django',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
    ],
    keywords=['vimeo', 'video'],
    install_requires=[
        'PyVimeo >=0.3.2, <0.4',
        'Django >=1.5, <1.10',
        'xxhash >=0.4.3, <0.5',
    ],
    test_suite='nose.collector',
)



"""
Django app for easy embeding Vimeo videos.
"""

VERSION = (0, 1, 7, 'dev')


def get_release():
    return '-'.join([get_version(), VERSION[-1]])


def get_version():
    """
    Returns only digit parts of version.
    """
    return '.'.join(str(i) for i in VERSION[:3])


__version__ = get_release()


import logging
import xxhash

from django.conf import settings
from django.core.cache import caches, InvalidCacheBackendError
from django.utils.functional import wraps

logger = logging.getLogger(__name__)


def cache_it(expires=None, key_func=None):
    if not expires:
        expires = getattr(settings, 'VIMEO_CACHE_EXPIRES', 300)
    cache = None
    try:
        cache_backend = getattr(settings, 'VIMEO_CACHE_BACKEND', None)
        cache = caches[cache_backend]
    except InvalidCacheBackendError as e:
        logger.warning('Vimeo cache "VIMEO_CACHE_BACKEND" disabled, reason: {}'.format(e))
    def wrap(f):
        @wraps(f)
        def wrapper(*args, **kwds):
            if not cache:
                return f(*args, **kwds)
            if key_func:
                key = 'django_vimeo_cache:{}'.format(key_func(*args, **kwds))
            else:
                key = 'django_vimeo_cache:' + f.__name__ + ':' +\
                      unicode(list(args) + list(sorted(kwds.items()))).encode('utf-8')
            key = xxhash.xxh64(key).hexdigest()
            value = cache.get(key)
            if value is None:
                value = f(*args, **kwds)
                cache.set(key, value, expires)
                value = cache.get(key)
                if value is None:
                    raise Exception('failed to fetch cached value, try again')
            return value
        return wrapper
    return wrap
